from typing import Optional

from silx.gui import qt

from tomwer.core.scan.scantype import ScanType
from tomwer.gui.control import datareacheractions
from tomwer.gui.control.datawatcher.controlwidget import ControlWidget


class ScanDiscoveryConfigWidget(qt.QGroupBox):
    sigScanTypeChanged = qt.Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setLayout(qt.QVBoxLayout())
        self._nxtomoQRB = qt.QRadioButton(ScanType.NX_TOMO.value, parent=self)
        self.layout().addWidget(self._nxtomoQRB)

        self._blissQRB = qt.QRadioButton(ScanType.BLISS.value, parent=self)
        self.layout().addWidget(self._blissQRB)

        self._specQRB = qt.QRadioButton(ScanType.SPEC.value, parent=self)
        self.layout().addWidget(self._specQRB)

        # set up
        self._nxtomoQRB.setChecked(True)
        # connect signal / slot
        self._nxtomoQRB.toggled.connect(self._configChanged)
        self._blissQRB.toggled.connect(self._configChanged)
        self._specQRB.toggled.connect(self._configChanged)

    def _configChanged(self, *args, **kwargs):
        self.sigScanTypeChanged.emit()

    def getScanType(self) -> ScanType:
        if self._specQRB.isChecked():
            return ScanType.SPEC
        elif self._blissQRB.isChecked():
            return ScanType.BLISS
        elif self._nxtomoQRB.isChecked():
            return ScanType.NX_TOMO
        else:
            raise NotImplementedError

    def setScanType(self, mode: ScanType):
        mode = ScanType.from_value(mode)
        if mode is ScanType.SPEC:
            self._specQRB.setChecked(True)
        elif mode is ScanType.BLISS:
            self._blissQRB.setChecked(True)
        elif mode is ScanType.NX_TOMO:
            self._nxtomoQRB.setChecked(True)
        else:
            raise NotImplementedError


class _MainWidget(qt.QWidget):
    sigStart = qt.Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self.controlWidget = ControlWidget()
        self.layout().addWidget(self.controlWidget)
        self.configWidget = ScanDiscoveryConfigWidget(parent=self)
        self.layout().addWidget(self.configWidget)

        # set up
        self.configWidget.setVisible(False)
        self.controlWidget.setVisible(False)
        self.controlWidget._qpbstartstop.setText("start discovery")
        # connect signal / slot
        self.controlWidget._qpbstartstop.released.connect(self._startDiscovery)
        # expose api
        self.sigScanTypeChanged = self.configWidget.sigScanTypeChanged

    def getConfigWindow(self):
        return self.configWidget

    def _startDiscovery(self):
        self.sigStart.emit()

    def getFolderObserved(self):
        return self.controlWidget._qteFolderSelected.text()

    def setFolderObserved(self, dir_):
        return self.controlWidget._qteFolderSelected.setText(dir_)

    def getLinuxFilePattern(self) -> Optional[str]:
        text = self.controlWidget._filterQLE.text()
        if text.replace(" ", "") == "":
            return None
        else:
            return text

    def setLinuxFilePattern(self, pattern: Optional[str]):
        if pattern is None:
            pattern = ""
        self.controlWidget._filterQLE.setText(pattern)

    def setSearchScanType(self, scan_type):
        self.configWidget.setScanType(scan_type)

    def getSearchScqnType(self):
        return self.configWidget.getScanType()


class DataDiscoveryWidget(qt.QMainWindow):
    sigTMStatusChanged = qt.Signal(str)
    """Signal emitted when the state changed"""
    sigScanReady = qt.Signal(object)
    """Signal emitted when a scan is considered as ready"""
    sigFolderObservedChanged = qt.Signal()
    """Signal emitted when the user change the observed folder"""
    sigObservationStart = qt.Signal()
    """Signal emitted when the observation starts"""
    sigObservationEnd = qt.Signal()
    """Signal emitted when the observation end"""
    sigFilterFileNamePatternChanged = qt.Signal(str)
    """Signal emut when the filter pattern change"""

    def __init__(self, parent=None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        self.setWindowFlags(qt.Qt.Widget)
        self._mainWidget = _MainWidget(parent=self)
        self.setCentralWidget(self.widget)
        # create toolbar
        toolbar = qt.QToolBar("", parent=self)
        toolbar.setIconSize(qt.QSize(32, 32))
        self._controlAction = datareacheractions.ControlAction(parent=self)
        self._configurationAction = datareacheractions.ConfigurationAction(parent=self)
        toolbar.addAction(self._controlAction)
        toolbar.addAction(self._configurationAction)

        self._actionGroup = qt.QActionGroup(self)
        self._actionGroup.addAction(self._controlAction)
        self._actionGroup.addAction(self._configurationAction)

        self.addToolBar(qt.Qt.LeftToolBarArea, toolbar)
        toolbar.setMovable(False)
        # connect signal / slot
        self._controlAction.toggled[bool].connect(
            self._mainWidget.controlWidget.setVisible
        )
        self._configurationAction.toggled[bool].connect(
            self._mainWidget.configWidget.setVisible
        )
        self._controlAction.setChecked(True)

    def getConfiguration(self) -> dict:
        return {
            "start_folder": self.widget.getFolderObserved(),
            "file_filter": self.widget.getLinuxFilePattern(),
            "scan_type_searched": self.widget.getConfigWindow().getScanType().value,
        }

    def setConfiguration(self, config: dict):
        if "start_folder" in config:
            self.widget.setFolderObserved(config["start_folder"])
        if "file_filter" in config:
            self.widget.setLinuxFilePattern(config["file_filter"])
        type_searched = config.get("scan_type_searched", None)
        if type_searched is not None:
            self.widget.getConfigWindow().setScanType(type_searched)

    @property
    def widget(self) -> qt.QWidget:
        return self._mainWidget



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/10/2016"

import functools
import logging

from orangewidget import gui, widget
from orangewidget.settings import Setting
from orangewidget.widget import Output

import tomwer.core.process.control.datawatcher.datawatcher
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datawatcher import DataWatcherWidget

from ..utils import WidgetLongProcessing

logger = logging.getLogger(__name__)


class DataWatcherOW(widget.OWBaseWidget, WidgetLongProcessing, openclass=True):
    """
    This widget is used to observe a selected folder and his sub-folders to
    detect if they are containing valid-finished acquisitions.
    """

    name = "scan watcher"
    id = "orangecontrib.widgets.tomwer.datawatcherwidget.DataWatcherOW"
    description = (
        "The widget will observe folder and sub folders of a given"
        " path and waiting for acquisition to be ended."
        " The widget will infinitely wait until an acquisition is "
        "ended. If an acquisition is ended then a signal "
        "containing the folder path is emitted."
    )
    icon = "icons/datawatcher.svg"
    priority = 12
    keywords = ["tomography", "file", "tomwer", "observer", "datawatcher"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    folderObserved = Setting(str())

    acquisitionMethod = Setting(tuple())

    linuxFilePatternSetting = Setting(str())

    DEFAULT_DIRECTORY = "/lbsram/data/visitor"

    ewokstaskclass = (
        tomwer.core.process.control.datawatcher.datawatcher.DataWatcherEwoksTask
    )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)
        bliss_scan = Output(
            name="bliss data", type=BlissScan, doc="bliss scan to be process"
        )

    def __init__(self, parent=None, displayAdvancement=True):
        """Simple class which will check advancement state of the acquisition
        for a specific folder

        :param parent: the parent widget
        """
        widget.OWBaseWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._widget = DataWatcherWidget(parent=self)
        self._widget.setFolderObserved(self.folderObserved)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # signal / slot connection
        self._widget.sigFolderObservedChanged.connect(self._updateSettings)
        self._widget.sigObservationModeChanged.connect(self._updateSettings)
        self._widget.sigObservationModeChanged.connect(self._setObsMethod)
        self._widget.sigFilterFileNamePatternChanged.connect(
            self._saveFilterLinuxPattern
        )
        self._widget.sigFilterFileNamePatternChanged.connect(
            self._widget.setLinuxFilePattern
        )
        self._widget.sigScanReady.connect(self._sendSignal)
        callback_start = functools.partial(self.processing_state, True, "watching on")
        self._widget.sigObservationStart.connect(callback_start)
        callback_end = functools.partial(self.processing_state, False, "watching off")
        self._widget.sigObservationEnd.connect(callback_end)
        self._loadSettings()

    def _loadSettings(self):
        if self.acquisitionMethod != tuple():
            self.widget.getConfigWindow().setMode(self.acquisitionMethod)
        if self.folderObserved != str():
            self.widget.setFolderObserved(self.folderObserved)
        if self.linuxFilePatternSetting != str():
            # for the GUI
            self.widget._filterQLE.setText(self.linuxFilePatternSetting)
            # for providing it to the processing thread. Bad design but this widget shouldn't be used that
            # much and modifying internals is not the priority.
            self.widget.setLinuxFilePattern(self.linuxFilePatternSetting)

    def _updateSettings(self):
        self.folderObserved = self.widget.getFolderObserved()
        mode = self.widget.getConfigWindow().getMode()
        if mode is not None:
            self.acquisitionMethod = mode
        self.linuxFilePatternSetting = self.widget.getLinuxFilePattern()

    def _saveFilterLinuxPattern(self, *args, **kwargs):
        linuxFilePattern = self.widget.getFilterLinuxFileNamePattern()
        if linuxFilePattern is None:
            self.linuxFilePatternSetting = ""
        else:
            self.linuxFilePatternSetting = linuxFilePattern

    def _setObsMethod(self, mode):
        self.widget.setObsMethod(mode)

    @property
    def widget(self):
        return self._widget

    @property
    def currentStatus(self):
        return self._widget.currentStatus

    @property
    def sigTMStatusChanged(self):
        return self._widget.sigTMStatusChanged

    def resetStatus(self):
        self._widget.resetStatus()

    def _sendSignal(self, scan):
        if scan is None:
            pass
        elif isinstance(scan, TomwerScanBase):
            self.Outputs.data.send(scan)
        elif isinstance(scan, BlissScan):
            self.Outputs.bliss_scan.send(scan)
        else:
            raise TypeError(
                f"output is expected to be a {TomwerScanBase} or a {BlissScan}"
            )

    def setFolderObserved(self, path):
        self.widget.setFolderObserved(path)

    def setObservation(self, b):
        self._widget.setObservation(b)

    def setTimeBreak(self, val):
        """
        Set the time break between two loop observation
        :param val: time (in sec)
        """
        self._widget.setWaitTimeBtwLoop(val)

    def startObservation(self):
        try:
            self.processing_info("processing")
        except Exception:
            pass
        self._widget.start()

    def stopObservation(self, succes=False):
        self.widget.stop(succes)
        try:
            self.Processing.clear()
        except Exception:
            pass

    def close(self):
        self.widget.close()
        super().close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import signal
import sys

import silx
from silx.gui import qt

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.samplemoved import SampleMovedWidget
from tomwer.gui.utils.splashscreen import getMainSplashScreen

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer samplemoved [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path", help="Data file to show (h5 file, edf files, spec files)"
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--entry",
        help="For Nexus files: entry in the master file",
        default=None,
    )
    parser.add_argument(
        "--use-opengl-plot",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    options = parser.parse_args(argv[1:])

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    widget = SampleMovedWidget()
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    scan = ScanFactory.create_scan_object(options.scan_path, options.entry)
    rawSlices = scan.get_proj_angle_url()
    widget.setImages(rawSlices)
    # TODO: set back the flat field correction
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



import pytest

from tomwer.core.process.reconstruction.axis.params import AxisRP
from tomwer.core.process.reconstruction.darkref.params import DKRFRP
from tomwer.core.process.reconstruction.paramsbase import _get_db_fromstr, _ReconsParam


def test_paramsbase():
    """
    Test _ReconsParam class
    """
    params = _ReconsParam()
    assert isinstance(params.all_params, list)


def test_get_db_fromstr():
    """
    test '_get_db_fromstr' function
    """
    assert _get_db_fromstr(vals="12") == 12.0
    assert _get_db_fromstr(vals="12.0") == 12.0
    assert _get_db_fromstr(vals="(12.0, )") == 12.0
    assert _get_db_fromstr(vals="(12.0, 13.5)") == (12.0, 13.5)
    assert _get_db_fromstr(vals="[12.0, 13.5]") == (12.0, 13.5)


def test_basic_DKRFRP():
    """
    dummy test for DKRFRP
    """
    recons_params = DKRFRP()
    ddict = recons_params.to_dict()
    DKRFRP.from_dict(ddict)
    assert isinstance(
        recons_params.to_unique_recons_set(),
        tuple,
    )


@pytest.mark.parametrize("param_klass", [DKRFRP, AxisRP])
def test_ReconsParams_from_dict(param_klass):
    """
    Test from_dict with upper and lower case keys
    """
    ref_dict = param_klass().to_dict()

    # Default case (upper)
    result = param_klass.from_dict(ref_dict.copy())
    assert result.to_dict() == ref_dict

    # Lower case
    lower_dict = {key.lower(): value for key, value in ref_dict.items()}
    result = param_klass.from_dict(lower_dict.copy())
    assert result.to_dict() == ref_dict



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016-2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/

# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "16/06/2021"


# import os
# import shutil
# import tempfile
# import unittest

# from tomwer.core.process.control.volumesymlink import VolumeSymbolicLinkTask
# from tomwer.core.scan.edfscan import EDFTomoScan
# from tomwer.core.utils.scanutils import MockEDF


# class TestVolumeSymbolicLink(unittest.TestCase):
#     """Simple test of the VolumeSymbolicLink"""

#     def setUp(self) -> None:
#         self._scan_folder = tempfile.mkdtemp()
#         self._output_folder = tempfile.mkdtemp()
#         MockEDF.fastMockAcquisition(self._scan_folder)
#         MockEDF.mockReconstruction(self._scan_folder, nRecons=2, nPagRecons=2)
#         self._scan = EDFTomoScan(self._scan_folder)
#         # here we simply estimate that the slice reconstruction will be
#         # volume reconstruction but the mechanism is the same.
#         self._scan.set_latest_vol_reconstructions(self._scan.get_reconstructions_urls())

#     def tearDown(self) -> None:
#         for folder in (self._scan_folder, self._output_folder):
#             shutil.rmtree(folder)

#     def test(self):
#         """simple test that the creation of a symbolic link works"""
#         vlink_process = VolumeSymbolicLinkTask(
#             inputs={
#                 "data": self._scan,
#                 "serialize_output_data": False,
#                 "output_type": "static",
#                 "output_folder": self._output_folder,
#             }
#         )
#         self.assertEqual(len(os.listdir(self._output_folder)), 0)
#         vlink_process.run()
#         self.assertEqual(len(os.listdir(self._output_folder)), 4)

#         self.assertTrue(isinstance(vlink_process.program_name(), str))
#         self.assertTrue(isinstance(vlink_process.program_version(), str))
#         self.assertTrue(isinstance(vlink_process.definition(), str))



try:
    __import__("pkg_resources").declare_namespace(__name__)
except ImportError:
    from pkgutil import extend_path

    __path__ = extend_path(__path__, __name__)
# orangecontrib is a namespace modules shared by multiple Orange add-on so it
# needs to declare namespace.



import os
import logging
from tomoscan.serie import Serie

from ewokscore.missing_data import is_missing_data

from tomwer.core.process.task import TaskWithProgress
from tomwer.core.scan.nxtomoscan import NXtomoScan

from nxtomo.application.nxtomo import NXtomo
from nxtomomill.utils.nexus import concatenate as nx_concatenate

_logger = logging.getLogger(__name__)


class ConcatenateNXtomoTask(
    TaskWithProgress,
    input_names=(
        "serie",
        "output_file",
        "output_entry",
        "overwrite",
    ),
    optional_input_names=(
        "progress",
        "serialize_output_data",
    ),
    output_names=("data",),
):
    """
    Task used to concatenate a list of NXtomo (NXtomoScan) into a single NXtomo
    """

    def run(self):
        scans_serie = self.inputs.serie
        if is_missing_data(scans_serie) or len(scans_serie) < 1:
            raise ValueError("expected at least one nxtomo to be concatenated")
        elif not isinstance(scans_serie, Serie):
            raise TypeError(f"serie is expected to be a Serie. Get {type(scans_serie)}")

        # format output location if necessary
        output_file = format_output_location(
            file_path=self.inputs.output_file, serie=scans_serie
        )

        # cast scan from NXtomoScan to NXtomo
        def cast_scan_to_nxtomo(obj: NXtomoScan):
            # convert NXtomoScan to NXtomo as we expect to get NXtomoScan
            if isinstance(obj, NXtomoScan):
                return NXtomo().load(
                    file_path=obj.master_file,
                    data_path=obj.entry,
                )
            else:
                raise TypeError(
                    f"nxtomos are supposed to be instances of {NXtomo} or {NXtomoScan}. Get {type(obj)}"
                )

        scans_serie = tuple([cast_scan_to_nxtomo(nxtomo) for nxtomo in scans_serie])

        # apply concatenation
        output_nxtomo = nx_concatenate(scans_serie)

        # dump result
        output_nxtomo.save(
            file_path=output_file,
            data_path=self.inputs.output_entry,
            overwrite=self.inputs.overwrite,
        )

        # cast back nxtomomill NXtomo to NXtomoScan (reference object for tomwer)
        scan = NXtomoScan(
            scan=output_file,
            entry=self.inputs.output_entry,
        )
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan


def format_output_location(file_path, serie: Serie):
    """
    format possible keys from the location like {scan_dir} or {scan_path}

    :param file_path:
    :param serie: serie of NXtomoScan
    :return:
    """
    if serie is None:
        _logger.warning("scan is !none, enable to format the nabu output location")

    for scan in serie:
        if not isinstance(scan, NXtomoScan):
            raise TypeError

    def get_common_path():
        if len(serie) == 0:
            return ""
        elif len(serie) == 1:
            return os.path.dirname(serie[0].master_file)
        else:
            return os.path.commonpath(
                [os.path.dirname(scan.master_file) for scan in serie]
            )

    keywords = {
        "common_path": get_common_path(),
    }

    # filter necessary keywords
    def get_necessary_keywords():
        import string

        formatter = string.Formatter()
        return [field for _, field, _, _ in formatter.parse(file_path) if field]

    requested_keywords = get_necessary_keywords()

    def keyword_needed(pair):
        keyword, _ = pair
        return keyword in requested_keywords

    keywords = dict(filter(keyword_needed, keywords.items()))
    file_path = file_path.format(**keywords)
    file_path = os.path.abspath(file_path)
    return file_path



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016-2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/

# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "16/11/2020"


# import logging
# import os

# from silx.utils.enum import Enum as _Enum
# from tomwer.core.utils.deprecation import deprecated_warning, deprecated
# from tomoscan.factory import Factory
# from tomoscan.volumebase import VolumeBase

# import tomwer.version
# from tomwer.core.process.task import Task
# from tomwer.core.scan.scanbase import TomwerScanBase
# from tomwer.core.scan.scanfactory import ScanFactory
# from tomwer.core.utils.scanutils import data_identifier_to_scan
# from tomwer.utils import docstring

# _logger = logging.getLogger(__name__)


# class OutputType(_Enum):
#     STATIC = "static"
#     ONE_LEVEL_UPPER = "../volume"


# def create_volume_symbolic_link(scan: TomwerScanBase, output_folder: str):
#     """
#     Create a symbolic link for each volume reconstructed of `scan`

#     :param TomwerScanBase scan:
#     :param str output_folder:
#     """
#     if scan.latest_vol_reconstructions is None:
#         _logger.info("No volume reconstructed found")
#         return
#     if not os.path.exists(output_folder):
#         os.makedirs(output_folder)
#     for volume_url in scan.latest_vol_reconstructions:
#         volume = Factory.create_tomo_object_from_identifier(volume_url)
#         assert isinstance(volume, VolumeBase)
#         volume_files = set(
#             list(volume.browse_data_files()) + list(volume.browse_metadata_files()),
#         )
#         for src_file_path in volume_files:
#             if src_file_path is not None:
#                 dst_file_path = os.path.join(
#                     output_folder, os.path.basename(src_file_path)
#                 )
#                 if os.path.exists(dst_file_path):
#                     _logger.info(
#                         "{} already exists. Cannot create a symbolic link on it"
#                     )
#                 else:
#                     print(
#                         "src_file is", src_file_path, "dst_file_path is", dst_file_path
#                     )
#                     os.symlink(src=src_file_path, dst=dst_file_path)


# class VolumeSymbolicLinkTask(
#     Task,
#     input_names=("data",),
#     output_names=("data",),
#     optional_input_names=("serialize_output_data", "output_type", "output_folder"),
# ):
#     """
#     Process class for volume symbolic link
#     """

#     def __init__(
#         self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
#     ):
#         super().__init__(
#             varinfo=varinfo,
#             inputs=inputs,
#             node_id=node_id,
#             node_attrs=node_attrs,
#             execinfo=execinfo,
#         )
#         self._output_type = OutputType.ONE_LEVEL_UPPER
#         self._output_folder = None

#     @docstring(Task.set_configuration)
#     @deprecated(
#         reason="ewoksification",
#         replacement="task.inputs.output_type and task.inputs.output_folder",
#         since_version="1.2",
#     )
#     def set_configuration(self, properties):
#         if "output_type" in properties:
#             self._output_type = OutputType.from_value(properties["output_type"])
#         if "output_folder" in properties:
#             self._output_folder = properties["output_folder"]

#     @deprecated(
#         reason="ewoksification",
#         replacement="task.inputs.output_type and task.inputs.output_folder",
#         since_version="1.2",
#     )
#     def get_configuration(self):
#         return super().get_configuration()

#     @docstring(Task.program_name)
#     @staticmethod
#     def program_name():
#         return "tomwer_volume_symlink"

#     @docstring(Task.program_version)
#     @staticmethod
#     def program_version():
#         return tomwer.version.version

#     @docstring(Task.run)
#     def run(self):
#         scan = data_identifier_to_scan(self.inputs.data)
#         if type(scan) is dict:
#             scan = ScanFactory.create_scan_object_frm_dict(scan)
#         else:
#             scan = scan
#         if scan is None:
#             return
#         if not isinstance(scan, TomwerScanBase):
#             raise TypeError(
#                 f"scan is expected to be a dict or an instance of TomwerScanBase. Not {type(scan)}"
#             )

#         output_type = OutputType.from_value(
#             self.get_input_value("output_type", self._output_type)
#         )
#         output_folder = self.get_input_value("output_folder", self._output_folder)
#         if output_type is OutputType.STATIC:
#             if output_folder is None:
#                 raise ValueError(
#                     "Manual setting of the output folder is "
#                     "requested but None is provided."
#                 )
#             else:
#                 if not os.path.isabs(output_folder):
#                     os.path.abspath(
#                         os.path.join(os.path.abspath(scan.path), output_folder)
#                     )
#         elif output_type is OutputType.ONE_LEVEL_UPPER:
#             output_folder = os.path.abspath(scan.path)
#             output_folder = os.path.abspath(
#                 os.path.join(output_folder, output_type.value)
#             )
#         else:
#             raise ValueError(f"output type {output_type.value} is not managed")
#         create_volume_symbolic_link(scan=scan, output_folder=output_folder)
#         if self.get_input_value("serialize_output_data", True):
#             self.outputs.data = scan.to_dict()
#         else:
#             self.outputs.data = scan

#     @docstring(Task.definition)
#     @staticmethod
#     def definition():
#         return "Create a symbolic link to the volume folder"


# class VolumeSymbolicLinkProcess(VolumeSymbolicLinkTask):
#     def __init__(
#         self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
#     ):
#         deprecated_warning(
#             name="tomwer.core.process.control.volumesymlink.VolumeSymbolicLinkProcess",
#             type_="class",
#             reason="improve readibility",
#             since_version="1.2",
#             replacement="VolumeSymbolicLinkTask",
#         )
#         super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)



from ewokscore.task import Task as EwoksTask

from tomwer.core.utils.scanutils import data_identifier_to_scan


class _VolumeSelectorPlaceHolder(
    EwoksTask, input_names=["volume"], output_names=["volume"]
):
    """
    task to select a volume or a set of volumes
    """

    def run(self):
        self.outputs.volume = data_identifier_to_scan(self.inputs.volume)



from orangecanvas.scheme import Scheme
from orangewidget.settings import SettingsPrinter
from orangewidget.workflow.widgetsscheme import WidgetManager
from orangewidget.workflow.widgetsscheme import (
    WidgetsSignalManager as _WidgetsSignalManager,
)
from silx.gui import qt


class WidgetsSignalManager(_WidgetsSignalManager):
    # skip signal compresion done in original orange version
    def compress_signals(self, signals):
        return signals


class WidgetsScheme(Scheme):
    def __init__(self, parent=None, title=None, description=None, env={}, **kwargs):
        super().__init__(parent, title, description, env=env, **kwargs)
        self.widget_manager = WidgetManager()
        self.signal_manager = WidgetsSignalManager(self)
        self.widget_manager.set_scheme(self)
        self.__report_view = None
        self.set_loop_flags(self.AllowLoops)

    def widget_for_node(self, node):
        """
        Return the OWBaseWidget instance for a `node`.
        """
        return self.widget_manager.widget_for_node(node)

    def node_for_widget(self, widget):
        """
        Return the SchemeNode instance for the `widget`.
        """
        return self.widget_manager.node_for_widget(widget)

    def sync_node_properties(self):
        """
        Sync the widget settings/properties with the SchemeNode.properties.
        Return True if there were any changes in the properties (i.e. if the
        new node.properties differ from the old value) and False otherwise.

        """
        changed = False
        for node in self.nodes:
            settings = self.widget_manager.widget_settings_for_node(node)
            if settings != node.properties:
                node.properties = settings
                changed = True
        return changed

    def show_report_view(self):
        return

    def has_report(self):
        """
        Does this workflow have an associated report

        Returns
        -------
        has_report: bool
        """
        return self.__report_view is not None

    def report_view(self):
        """
        Return a OWReport instance used by the workflow.

        Returns
        -------
        report : OWReport
        """
        return None

    def set_report_view(self, view):
        """
        Set the designated OWReport view for this workflow.

        Parameters
        ----------
        view :
        """
        self.__report_view = None

    def dump_settings(self, node):
        widget = self.widget_for_node(node)

        pp = SettingsPrinter(indent=4)
        pp.pprint(widget.settingsHandler.pack_data(widget))

    def event(self, event):
        if event.type() == qt.QEvent.Close:
            if self.__report_view is not None:
                self.__report_view.close()
            self.signal_manager.stop()
        return super().event(event)



import weakref
from typing import Optional
from silx.gui import qt
from silx.gui.plot import Plot2D
from silx.gui.colors import Colormap

from tomwer.gui import icons
from tomwer.gui.settings import Y_AXIS_DOWNWARD
from tomwer.gui.stitching.stitchandbackground import StitchAndBackgroundAlphaMixIn


class FullScreenPlot2D(Plot2D):
    """
    Window which goal is to display a single frame into full screen

    :param master_colormap: optional colormap that could redefine values of the
    current plot colormap
    """

    def __init__(self, parent=None, backend=None, master_colormap=None):
        super().__init__(parent, backend)
        if master_colormap is None:
            self.__master_colormap = master_colormap
        else:
            assert isinstance(master_colormap, Colormap)
            self.__master_colormap = weakref.ref(master_colormap)

        self.getPositionInfoWidget().hide()
        self.setYAxisInverted(Y_AXIS_DOWNWARD)

        self.setAxesDisplayed(False)
        self.getColorBarWidget().hide()
        self.setWindowIcon(icons.getQIcon("tomwer"))

        # add toolbar to ease exit
        self.exit_toolbar = qt.QToolBar("exit", self)
        # add a spacer for conveniance
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Preferred)
        self.exit_toolbar.addWidget(spacer)
        self.addToolBar(self.exit_toolbar)
        # add exit button wen displayed full screen
        exit_icon = icons.getQIcon("exit")
        self._quitAction = qt.QAction(exit_icon, "exit", self)
        self.exit_toolbar.addAction(self._quitAction)

        if self.master_colormap is not None:
            master_colormap.sigChanged.connect(self._updateColormapFromMaster)

        self._updateColormapFromMaster()

        # connect signal / slot
        self._quitAction.triggered.connect(self.close)

    @property
    def master_colormap(self) -> Optional[Colormap]:
        if self.__master_colormap is None:
            return None
        else:
            return self.__master_colormap()

    def _updateColormapFromMaster(self):
        if self.master_colormap is not None:
            current_colormap = self.getDefaultColormap()
            current_colormap.setFromColormap(self.master_colormap)


class FullScreenStitching(FullScreenPlot2D, StitchAndBackgroundAlphaMixIn):
    def __init__(
        self,
        stitching_img,
        background_img,
        parent=None,
        backend=None,
        master_colormap=None,
    ):
        super().__init__(parent, backend, master_colormap)
        self._stitched_image = stitching_img
        self._composition_background = background_img
        self._update()

        self._backgroundToolbar = qt.QToolBar("background")
        self._backgroundToolbar.addAction(self._backGroundAction)
        self.insertToolBar(self.toolBar(), self._backgroundToolbar)
        self._backgroundToolbar.addWidget(self._alphaChannelWidget)

        # set up
        self.setKeepDataAspectRatio(True)

    def _updateAlphaImg(self, *args, **kwargs):
        alpha_img = self.getImage(self.LEGEND_STITCHED_FRAME)
        if alpha_img is not None:
            alpha_img.setAlpha(self.getStitchedImgAlpha())

    def _updateAlphaBackgroundImg(self, *args, **kwargs):
        background_img = self.getImage(self.LEGEND_BACKGROUND)
        if background_img is not None:
            background_img.setAlpha(self.getBackgroundImgAlpha())

    @property
    def mother_plot(self) -> Optional[Plot2D]:
        if self._mother_plot is None:
            return None
        else:
            return self._mother_plot()



import logging
import weakref
from typing import Optional
from functools import lru_cache as cache

import h5py
import numpy
from silx.gui import qt
from silx.io.utils import h5py_read_dataset
from silx.io.dictdump import dicttonx

from tomoscan.io import HDF5File
from nxtomo.paths.nxtomo import get_paths as get_nexus_paths
from nxtomo.nxobject.nxdetector import ImageKey
from nxtomo.nxobject.nxdetector import FOV
from nxtomo.nxobject.nxtransformations import NXtransformations
from nxtomo.utils.transformation import (
    build_matrix,
    UDDetTransformation,
    LRDetTransformation,
)

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.utils.scandescription import ScanNameLabelAndShape
from tomwer.gui.utils.unitsystem import MetricEntry
from tomwer.gui.edit.nxtomowarmer import NXtomoProxyWarmer

_logger = logging.getLogger(__name__)


class NXtomoEditorDialog(qt.QDialog):
    def __init__(self, parent=None, hide_lockers=True) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self.mainWidget = NXtomoEditor(parent=self, hide_lockers=hide_lockers)
        self.layout().addWidget(self.mainWidget)
        self._warmer = NXtomoProxyWarmer(parent=self)
        self.layout().addWidget(self._warmer)

        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("validate")
        self.layout().addWidget(self._buttons)

    # expose API
    def setScan(self, scan):
        self.mainWidget.setScan(scan)
        self._warmer.setScan(scan)

    def overwriteNXtomo(self):
        self.mainWidget.overwriteNXtomo()

    def hasLockField(self) -> bool:
        return self.mainWidget.hasLockField()

    def getConfiguration(self) -> dict:
        return self.mainWidget.getConfiguration()

    def setConfiguration(self, config):
        self.mainWidget.setConfiguration(config)


class NXtomoEditor(qt.QWidget):
    """
    class to edit parameter of a NXtomo.
    The preliminary goal is to let the user define pixel / voxel position and x and z positions
    in order to simplify stitching down the line

    As energy and scan range was also often requested this part is also editable (user bonus ^^)
    """

    sigEditingFinished = qt.Signal()

    def __init__(self, parent=None, hide_lockers=True):
        super().__init__(parent)
        self._editableWidgets = []
        self._lockerPBs = []
        # list of all lockers
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._scanInfoQLE = ScanNameLabelAndShape(parent=self)
        self.layout().addWidget(self._scanInfoQLE)

        # nxtomo tree
        self._tree = qt.QTreeWidget(self)
        if hide_lockers:
            self._tree.setColumnCount(2)
            self._tree.setHeaderLabels(("entry", "value"))
        else:
            self._tree.setColumnCount(3)
            self._tree.setHeaderLabels(("entry", "value", "lockers"))
            self._tree.header().setStretchLastSection(False)
            self._tree.setColumnWidth(2, 20)
            self._tree.header().setSectionResizeMode(1, qt.QHeaderView.Stretch)
        self.layout().addWidget(self._tree)

        # 1: instrument
        self._instrumentQTWI = qt.QTreeWidgetItem(self._tree)
        self._instrumentQTWI.setText(0, "instrument")
        # handle energy
        self._beamQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._beamQTWI.setText(0, "beam")
        self._energyQTWI = qt.QTreeWidgetItem(self._beamQTWI)
        self._energyQTWI.setText(0, "energy (keV)")
        self._energyEntry = EnergyEntry("", self)
        self._energyEntry.setPlaceholderText("energy in kev")
        self._tree.setItemWidget(self._energyQTWI, 1, self._energyEntry)
        self._editableWidgets.append(self._energyEntry)
        self._energyLockerLB = PadlockButton(self)
        self._energyLockerLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._energyLockerLB)
        self._tree.setItemWidget(self._energyQTWI, 2, self._energyLockerLB)

        # 1.1 detector
        self._detectorQTWI = qt.QTreeWidgetItem(self._instrumentQTWI)
        self._detectorQTWI.setText(0, "detector")
        ## pixel size
        self._xPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._xPixelSizeQTWI.setText(0, "x pixel size")
        self._xPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._xPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(self._xPixelSizeQTWI, 1, self._xPixelSizeMetricEntry)
        self._editableWidgets.append(self._xPixelSizeMetricEntry)
        self._xPixelSizeLB = PadlockButton(self)
        self._xPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xPixelSizeLB)
        self._tree.setItemWidget(self._xPixelSizeQTWI, 2, self._xPixelSizeLB)

        self._yPixelSizeQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._yPixelSizeQTWI.setText(0, "y pixel size")
        self._yPixelSizeMetricEntry = MetricEntry("", parent=self)
        self._yPixelSizeMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(self._yPixelSizeQTWI, 1, self._yPixelSizeMetricEntry)
        self._editableWidgets.append(self._yPixelSizeMetricEntry)
        self._yPixelSizeLB = PadlockButton(self)
        self._yPixelSizeLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._yPixelSizeLB)
        self._tree.setItemWidget(self._yPixelSizeQTWI, 2, self._yPixelSizeLB)

        ## distance
        self._sampleDetectorDistanceQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._sampleDetectorDistanceQTWI.setText(0, "distance")
        self._distanceMetricEntry = MetricEntry("", parent=self)
        self._distanceMetricEntry.layout().setContentsMargins(2, 2, 2, 2)
        self._tree.setItemWidget(
            self._sampleDetectorDistanceQTWI, 1, self._distanceMetricEntry
        )
        self._editableWidgets.append(self._distanceMetricEntry)
        self._distanceLB = PadlockButton(self)
        self._distanceLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._distanceLB)
        self._tree.setItemWidget(self._sampleDetectorDistanceQTWI, 2, self._distanceLB)

        ## field of view
        self._fieldOfViewQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._fieldOfViewQTWI.setText(0, "field of view")
        self._fieldOfViewCB = qt.QComboBox(self)
        for value in FOV.values():
            self._fieldOfViewCB.addItem(value)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 1, self._fieldOfViewCB)
        self._editableWidgets.append(self._fieldOfViewCB)
        self._fieldOfViewLB = PadlockButton(self)
        self._fieldOfViewLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._fieldOfViewLB)
        self._tree.setItemWidget(self._fieldOfViewQTWI, 2, self._fieldOfViewLB)

        ## x flipped
        self._xFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._xFlippedQTWI.setText(0, "x flipped")
        self._xFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._xFlippedQTWI, 1, self._xFlippedCB)
        self._editableWidgets.append(self._xFlippedCB)
        self._xFlippedLB = PadlockButton(self)
        self._xFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._xFlippedLB)
        self._tree.setItemWidget(self._xFlippedQTWI, 2, self._xFlippedLB)
        ## y flipped
        self._yFlippedQTWI = qt.QTreeWidgetItem(self._detectorQTWI)
        self._yFlippedQTWI.setText(0, "y flipped")
        self._yFlippedCB = qt.QCheckBox("", self)
        self._tree.setItemWidget(self._yFlippedQTWI, 1, self._yFlippedCB)
        self._editableWidgets.append(self._yFlippedCB)
        self._yFlippedLB = PadlockButton(self)
        self._yFlippedLB.setMaximumSize(30, 30)
        self._lockerPBs.append(self._yFlippedLB)
        self._tree.setItemWidget(self._yFlippedQTWI, 2, self._yFlippedLB)
        # 2: sample
        self._sampleQTWI = qt.QTreeWidgetItem(self._tree)
        self._sampleQTWI.setText(0, "sample")
        ## x translation
        self._xTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._xTranslationQTWI.setText(0, "x translation")
        self._xTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._xTranslationQTWI, 1, self._xTranslationQLE)
        self._editableWidgets.append(self._xTranslationQLE)

        ## z translation
        self._zTranslationQTWI = qt.QTreeWidgetItem(self._sampleQTWI)
        self._zTranslationQTWI.setText(0, "z translation")
        self._zTranslationQLE = _TranslationMetricEntry(name="", parent=self)
        self._tree.setItemWidget(self._zTranslationQTWI, 1, self._zTranslationQLE)
        self._editableWidgets.append(self._zTranslationQLE)

        # set up
        self._instrumentQTWI.setExpanded(True)
        self._sampleQTWI.setExpanded(True)
        self._beamQTWI.setExpanded(True)
        self._detectorQTWI.setExpanded(True)
        self.hideLockers(hide_lockers)

        # connect signal / slot
        self._energyEntry.editingFinished.connect(self._editingFinished)
        self._energyLockerLB.toggled.connect(self._editingFinished)
        self._xPixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._xPixelSizeLB.toggled.connect(self._editingFinished)
        self._yPixelSizeMetricEntry.editingFinished.connect(self._editingFinished)
        self._yPixelSizeLB.toggled.connect(self._editingFinished)
        self._distanceMetricEntry.editingFinished.connect(self._editingFinished)
        self._distanceLB.toggled.connect(self._editingFinished)
        self._fieldOfViewCB.currentIndexChanged.connect(self._editingFinished)
        self._fieldOfViewLB.toggled.connect(self._editingFinished)
        self._xFlippedCB.toggled.connect(self._editingFinished)
        self._xFlippedLB.toggled.connect(self._editingFinished)
        self._yFlippedCB.toggled.connect(self._editingFinished)
        self._yFlippedLB.toggled.connect(self._editingFinished)

    def _editingFinished(self, *args, **kwargs):
        self.sigEditingFinished.emit()

    def hasLockField(self) -> bool:
        """return True if the widget has at least one lock field"""
        return True in [locker.isLocked() for locker in self._lockerPBs]

    def hideLockers(self, hide: bool):
        for locker in self._lockerPBs:
            locker.setVisible(not hide)

    def getEditableWidgets(self):
        return self._editableWidgets

    def setScan(self, scan):
        if scan is None:
            self._scan = scan
        elif not isinstance(scan, NXtomoScan):
            raise TypeError(
                f"{scan} is expected to be an instance of {NXtomoScan}. Not {type(scan)}"
            )
        else:
            self._scan = weakref.ref(scan)
        self._scanInfoQLE.setScan(scan)
        # scan will only be read and not kept
        self.update_tree()

    def getScan(self):
        if self._scan is None or self._scan() is None:
            return None
        else:
            return self._scan()

    def update_tree(self):
        if self.getScan() is not None:
            for fct in (
                self._updateInstrument,
                self._updateSample,
            ):
                try:
                    fct()
                except Exception as e:
                    _logger.info(e)
            self._tree.resizeColumnToContents(0)

    def _updateInstrument(self):
        scan = self.getScan()
        if scan is None:
            return
        else:
            self._updateEnergy(scan=scan)
            self._updatePixelSize(scan=scan)
            self._updateFlipped(scan=scan)
            self._updateFieldOfView(scan=scan)
            self._updateDistance(scan=scan)

    def _updateSample(self):
        scan = self.getScan()
        if scan is None:
            return
        else:
            self._updateTranslations(scan=scan)

    def _updateTranslations(self, scan: NXtomoScan):
        assert isinstance(scan, NXtomoScan)

        # note: for now and in order to allow edition we expect to have at most a unique value. Will fail for helicoidal
        def reduce(values):
            if values is None:
                return None
            values = numpy.array(values)
            values = numpy.unique(
                values[scan.image_key_control == ImageKey.PROJECTION.value]
            )
            if values.size == 1:
                return values[0]
            elif values.size == 0:
                return None
            else:
                return f"{values[0]} ... {values[-1]}"

        x_translation = reduce(scan.x_translation)
        z_translation = reduce(scan.z_translation)
        self._xTranslationQLE.setValue(x_translation)
        self._zTranslationQLE.setValue(z_translation)

    def _updateFieldOfView(self, scan):
        if not self._fieldOfViewLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            idx = self._fieldOfViewCB.findText(FOV.from_value(scan.field_of_view).value)
            if idx > 0:
                self._fieldOfViewCB.setCurrentIndex(idx)

    @staticmethod
    @cache(maxsize=None)
    def _get_UD_flip_matrix():
        return UDDetTransformation().as_matrix()

    @staticmethod
    @cache(maxsize=None)
    def _get_LR_flip_matrix():
        return LRDetTransformation().as_matrix()

    @staticmethod
    @cache(maxsize=None)
    def _get_UD_AND_LR_flip_matrix():
        return numpy.matmul(
            NXtomoEditor._get_UD_flip_matrix(),
            NXtomoEditor._get_LR_flip_matrix(),
        )

    def _updateFlipped(self, scan):
        transformations = list(scan.get_detector_transformations(tuple()))
        transformation_matrix_det_space = build_matrix(transformations)
        if transformation_matrix_det_space is None or numpy.allclose(
            transformation_matrix_det_space, numpy.identity(3)
        ):
            flip_ud = False
            flip_lr = False
        elif numpy.array_equal(
            transformation_matrix_det_space, NXtomoEditor._get_UD_flip_matrix()
        ):
            flip_ud = True
            flip_lr = False
        elif numpy.allclose(
            transformation_matrix_det_space, NXtomoEditor._get_LR_flip_matrix()
        ):
            flip_ud = False
            flip_lr = True
        elif numpy.allclose(
            transformation_matrix_det_space, NXtomoEditor._get_UD_AND_LR_flip_matrix()
        ):
            flip_ud = True
            flip_lr = True
        else:
            flip_ud = None
            flip_lr = None
            _logger.warning(
                "detector transformations provided not handled... For now only handle up-down flip as left-right flip"
            )
        if (not self._xFlippedLB.isLocked()) and flip_lr is not None:
            self._xFlippedCB.setChecked(flip_lr)
        if (not self._yFlippedLB.isLocked()) and flip_ud is not None:
            self._yFlippedCB.setChecked(flip_ud)

    def _updateDistance(self, scan):
        if not self._distanceLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            self._distanceMetricEntry.setValue(scan.distance)

    def _updateEnergy(self, scan):
        assert isinstance(scan, NXtomoScan)
        if not self._energyLockerLB.isLocked():
            # if in ''auto mode: we want to overwrite the NXtomo existing value by the one of the GUI
            energy = scan.energy
            self._energyEntry.setValue(energy)

    def _updatePixelSize(self, scan):
        assert isinstance(scan, NXtomoScan)
        if not self._xPixelSizeLB.isLocked():
            x_pixel_size = scan.x_pixel_size
            self._xPixelSizeMetricEntry.setValue(x_pixel_size)
        if not self._yPixelSizeLB.isLocked():
            y_pixel_size = scan.y_pixel_size
            self._yPixelSizeMetricEntry.setValue(y_pixel_size)

    def getConfiguration(self) -> dict:
        """
        Return a dict with field full name as key
        and a tuple as value (field_value, is_locked)

        limitation: for now sample position are not handled because this is a 'corner case' for now
        """
        return {
            "instrument.beam.energy": (
                self._energyEntry.getValue(),
                self._energyLockerLB.isLocked(),
            ),
            "instrument.detector.x_pixel_size": (
                self._xPixelSizeMetricEntry.getValue(),
                self._xPixelSizeLB.isLocked(),
            ),
            "instrument.detector.y_pixel_size": (
                self._yPixelSizeMetricEntry.getValue(),
                self._yPixelSizeLB.isLocked(),
            ),
            "instrument.detector.distance": (
                self._distanceMetricEntry.getValue(),
                self._distanceLB.isLocked(),
            ),
            "instrument.detector.field_of_view": (
                self._fieldOfViewCB.currentText(),
                self._fieldOfViewLB.isChecked(),
            ),
            "instrument.detector.x_flipped": (
                self._xFlippedCB.isChecked(),
                self._xFlippedLB.isChecked(),
            ),
            "instrument.detector.y_flipped": (
                self._yFlippedCB.isChecked(),
                self._yFlippedLB.isChecked(),
            ),
        }

    def setConfiguration(self, config: dict):
        energy = config.get("instrument.beam.energy", None)
        if energy is not None:
            energy, energy_locked = energy
            self._energyEntry.setValue(energy)
            self._energyLockerLB.setLock(energy_locked)

        x_pixel_size = config.get("instrument.detector.x_pixel_size", None)
        if x_pixel_size is not None:
            x_pixel_size, x_pixel_size_locked = x_pixel_size
            self._xPixelSizeMetricEntry.setValue(x_pixel_size)
            self._xPixelSizeLB.setLock(x_pixel_size_locked)

        y_pixel_size = config.get("instrument.detector.y_pixel_size", None)
        if y_pixel_size is not None:
            y_pixel_size, y_pixel_size_locked = y_pixel_size
            self._yPixelSizeMetricEntry.setValue(y_pixel_size)
            self._yPixelSizeLB.setLock(y_pixel_size_locked)

        detector_sample_distance = config.get("instrument.detector.distance", None)
        if detector_sample_distance is not None:
            detector_sample_distance, distance_locked = detector_sample_distance
            self._distanceMetricEntry.setValue(detector_sample_distance)
            self._distanceLB.setLock(x_pixel_size_locked)

        field_of_view = config.get("instrument.detector.field_of_view", None)
        if field_of_view is not None:
            field_of_view, field_of_view_locked = field_of_view
            self._fieldOfViewCB.setCurrentText(field_of_view)
            self._fieldOfViewLB.setLock(field_of_view_locked)

        x_flipped = config.get("instrument.detector.x_flipped", None)
        if x_flipped is not None:
            x_flipped, x_flipped_locked = x_flipped
            x_flipped = x_flipped in (True, "True", "true")
            self._xFlippedCB.setChecked(x_flipped)
            self._xFlippedLB.setLock(x_flipped_locked)

        y_flipped = config.get("instrument.detector.y_flipped", None)
        if y_flipped is not None:
            y_flipped, y_flipped_locked = y_flipped
            y_flipped = y_flipped in (True, "True", "true")
            self._yFlippedCB.setChecked(y_flipped)
            self._yFlippedLB.setLock(y_flipped_locked)

    def clear(self):
        self._tree.clear()

    def overwriteNXtomo(self):
        """overwrite data on disk"""
        scan = self.getScan()
        if scan is None:
            _logger.warning("no scan found to be saved")
            return
        nexus_paths = get_nexus_paths(scan.nexus_version)
        assert isinstance(scan, NXtomoScan)
        with HDF5File(scan.master_file, mode="a") as h5f:
            entry = h5f[scan.entry]
            # overwrite energy
            energy = self._energyEntry.getValue()
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.ENERGY_PATH,
                value=energy,
                name="energy",
                expected_type=float,
                units="kev",
            )
            # overwrite x pixel size
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.X_PIXEL_SIZE_PATH,
                value=self._xPixelSizeMetricEntry.getValue(),
                name="x pixel size",
                expected_type=float,
                units="m",
            )
            # overwrite y pixel size
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.Y_PIXEL_SIZE_PATH,
                value=self._yPixelSizeMetricEntry.getValue(),
                name="y pixel size",
                expected_type=float,
                units="m",
            )
            n_frames = len(scan.image_key_control)

            # overwrite x translation
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.X_TRANS_PATH,
                value=self._xTranslationQLE.getValue(),
                name="x translation",
                expected_type=float,
                n_value=n_frames,
                units="m",
            )
            # overwrite z translation
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.Z_TRANS_PATH,
                value=self._zTranslationQLE.getValue(),
                name="z translation",
                expected_type=float,
                n_value=n_frames,
                units="m",
            )
            # overwrite sample detector distance
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.DISTANCE_PATH,
                value=self._distanceMetricEntry.getValue(),
                name="z translation",
                expected_type=float,
                units="m",
            )
            # overwrite FOV
            self.__write_to_file(
                entry=entry,
                path=nexus_paths.FOV_PATH,
                value=self._fieldOfViewCB.currentText(),
                name="field of view",
                expected_type=str,
            )

            # solve NXtranformations
            nx_transformations = NXtransformations()
            if self._xFlippedCB.isChecked():
                nx_transformations.add_transformation(LRDetTransformation())
            if self._yFlippedCB.isChecked():
                nx_transformations.add_transformation(UDDetTransformation())

            nx_dict = nx_transformations.to_nx_dict(
                nexus_path_version=scan.nexus_version,
                data_path="/".join(
                    (
                        nexus_paths.INSTRUMENT_PATH,
                        nexus_paths.nx_instrument_paths.DETECTOR_PATH,
                    )
                ),
                solve_empty_dependency=True,
            )
            if nexus_paths.nx_detector_paths.NX_TRANSFORMATIONS is not None:
                # old NXtomo are not handling NX_TRANSFORMATIONS
                detector_transformation_path = "/".join(
                    (
                        nexus_paths.INSTRUMENT_PATH,
                        nexus_paths.nx_instrument_paths.DETECTOR_PATH,
                        nexus_paths.nx_detector_paths.NX_TRANSFORMATIONS,
                    ),
                )
                if detector_transformation_path in entry:
                    del entry[detector_transformation_path]

                detector_transformation_path = "/".join(
                    (scan.entry, detector_transformation_path)
                )
            else:
                _logger.debug(
                    "Old version of NXtomo found. No information about transformation will be saved"
                )
                detector_transformation_path = None

        if detector_transformation_path is not None:
            dicttonx(
                nx_dict,
                h5file=scan.master_file,
                h5path=detector_transformation_path,
                update_mode="replace",
                mode="a",
            )

        # clear caches to make sure all modifications will be considered
        scan.clear_caches()
        scan.clear_frames_caches()

    @staticmethod
    def _newValueIsExistingValue(dataset: h5py.Dataset, new_value, units):
        """
        return true if the given value is same as the one stored
        """
        current_value = h5py_read_dataset(dataset)
        attrs = dataset.attrs
        current_unit = attrs.get("units", attrs.get("unit", None))
        if units != current_unit:
            # if the unit is not the same, eithen if the value is the same we will overwrite it
            return False
        else:
            if isinstance(new_value, numpy.ndarray) and isinstance(
                current_value, numpy.ndarray
            ):
                return numpy.array_equal(new_value, current_value)
            elif numpy.isscalar(current_value) and numpy.isscalar(new_value):
                return current_value == new_value
            else:
                return False

    @staticmethod
    def __write_to_file(
        entry: h5py.Group, path: str, value, name, expected_type, n_value=1, units=None
    ):
        if path is None:
            # if the path does not exists (no handled by this version of nexus for example)
            return

        # try to cast the value
        if isinstance(value, str):
            value = value.replace(" ", "")
            if value.lower() == "none" or "..." in value:
                # if value is not defined or is an array not overwrite by the user (case of the ... )
                return
        elif value is None:
            pass
        else:
            try:
                value = expected_type(value)
            except (ValueError, TypeError) as e:
                _logger.error(f"Fail to overwrite {name} of {entry.name}. Error is {e}")
                return

        if path in entry:
            if NXtomoEditor._newValueIsExistingValue(
                dataset=entry[path], new_value=value, units=units
            ):
                # if no need to overwrite
                return
            else:
                del entry[path]
        if value is None:
            return
        elif n_value == 1:
            entry[path] = value
        else:
            entry[path] = numpy.array([value] * n_value)
        if units is not None:
            entry[path].attrs["units"] = units


class _TranslationMetricEntry(MetricEntry):
    LOADED_ARRAY = "loaded array"

    class TranslationValidator(qt.QDoubleValidator):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.setNotation(qt.QDoubleValidator.ScientificNotation)

        def validate(self, a0: str, a1: int):
            if "..." in a0:
                return (qt.QDoubleValidator.Acceptable, a0, a1)
            else:
                return super().validate(a0, a1)

    def __init__(self, name, default_unit="m", parent=None):
        super().__init__(name, default_unit=default_unit, parent=parent)
        self._qlePixelSize.setValidator(self.TranslationValidator(self))

    def getValue(self):
        """

        :return: the value in meter
        :rtype: float
        """
        if "..." in self._qlePixelSize.text():
            # in this case this is the representation of an array, we don;t wan't to overwrite it
            return self.LOADED_ARRAY
        if self._qlePixelSize.text() in ("unknown", ""):
            return None
        else:
            return float(self._qlePixelSize.text()) * self.getCurrentUnit()


class EnergyEntry(qt.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setValidator(MetricEntry.DoubleValidator())

    def setValue(self, a0):
        if a0 is None:
            a0 = "unknown"
        else:
            a0 = str(a0)
        super().setText(a0)

    def getValue(self) -> Optional[float]:
        txt = self.text().replace(" ", "")
        if txt in ("unknown", ""):
            return None
        else:
            return float(txt)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2020"


import os
import shutil
import tempfile
import unittest
from glob import glob

from tomwer.core.process.reconstruction.axis import AxisRP
from tomwer.core.process.reconstruction.nabu import nabuslices as nabu
from tomwer.core.process.reconstruction.nabu import settings as nabu_settings
from tomwer.core.process.reconstruction.nabu import utils as nabuutils
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSliceMode
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import MockEDF
from tomwer.tests.datasets import TomwerCIDatasets


class TestNabuDescUtils(unittest.TestCase):
    """Test the util function for creating the description of nabu"""

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.top_dir = tempfile.mkdtemp()
        self.dataset = "scan_3_"
        data_test_dir = TomwerCIDatasets.get_dataset(f"edf_datasets/{self.dataset}")
        scan_folder = os.path.join(self.top_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = ScanFactory.create_scan_object(scan_folder)

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def testResourcesDescriptor(self):
        nabuutils.get_nabu_resources_desc(scan=self.scan, method="local")
        nabuutils.get_nabu_resources_desc(scan=self.scan, method="slurm")

    def testDatasetDescriptor(self):
        self.scan.get_nabu_dataset_info()


class TestNabuIni(unittest.TestCase):
    """Test creation of nabu .cfg files according to tomwer configuration"""

    def setUp(self) -> None:
        self.dataset = "scan_3_"
        self._root_dir = tempfile.mkdtemp()
        data_test_dir = TomwerCIDatasets.get_dataset("edf_datasets/scan_3_")
        scan_folder = os.path.join(self._root_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = EDFTomoScan(scan_folder)
        self._nabu_cfg_folder = os.path.join(
            self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
        )
        [os.remove(cfg_file) for cfg_file in glob(self.scan.path + os.sep + "*.cfg")]
        if os.path.exists(self._nabu_cfg_folder):
            shutil.rmtree(self._nabu_cfg_folder)

        assert len(self.getNabuIniFiles()) == 0
        # a random configuration
        self.config = get_nabu_default_config(output_dir=self.scan.path)

    def tearDown(self) -> None:
        shutil.rmtree(self._root_dir)

    def testNoSlices(self):
        """If no slices selected, in this case we will only create a .cfg
        for the volume
        """
        self.config["tomwer_slices"] = None
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            1,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 1)
        # check volume and slice .cfg files for nabu reconstruction are here
        ini_file = os.path.basename(self.scan.path) + ".cfg"
        ini_file = os.path.join(self._nabu_cfg_folder, ini_file)
        self.assertTrue(ini_file in nabu_files)

    def testSingleSlice(self):
        """Test the behavior if we only request one slice without Paganin"""
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            2,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 2)
        # check volume and slice .cfg files for nabu reconstruction are here
        for file_name in (
            os.path.basename(self.scan.path) + ".cfg",
            os.path.basename(self.scan.path) + "slice_000001_plane_XY" + ".cfg",
        ):
            with self.subTest(file_name=file_name):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, file_name) in nabu_files
                )

    def testSinglePagSlice(self):
        """Test the behavior if we only request one Paganin slice"""
        self.config["phase"]["method"] = "Paganin"
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["delta_beta"] = "100.0"
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            2,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 2)
        # check volume and slice .cfg files for nabu reconstruction are here
        for file_name in (
            os.path.basename(self.scan.path) + "pag_db0100.cfg",
            os.path.basename(self.scan.path)
            + "slice_pag_000001_db0100_plane_XY"
            + ".cfg",
        ):
            with self.subTest(file_name=file_name):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, file_name) in nabu_files
                )

    def testSeveralSlices(self):
        """Test the behavior if we request several slices without Paganin"""
        slices = (1, 4, 12)
        self.config["tomwer_slices"] = str(slices)
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            len(slices) + 1,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(slices) + 1)
        # check volume and slice .ini files for nabu reconstruction are here
        ini_files = [os.path.basename(self.scan.path) + ".cfg"]
        for slice_index in slices:
            ini_files.append(
                os.path.basename(self.scan.path)
                + "slice_"
                + str(slice_index).zfill(6)
                + "_plane_XY"
                + ".cfg"
            )

        print("nabu_files are", nabu_files)
        for ini_file in ini_files:
            print("check ini_file", ini_file)
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def testSeveralPags(self):
        """behavior when several paganin slices are requested"""
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["method"] = "Paganin"
        pag_dbs = [200.0, 300.0, 600.0, 700.0]
        self.config["phase"]["delta_beta"] = str(pag_dbs)
        sub_configs = nabu.interpret_tomwer_configuration(
            config=self.config, scan=self.scan
        )
        self.assertEqual(len(sub_configs), len(pag_dbs) * 2)
        # check that each dbs and slices values are in sub configurations
        existing_dbs = [config[0]["phase"]["delta_beta"] for config in sub_configs]
        existing_dbs = set(existing_dbs)
        for pag_db in pag_dbs:
            with self.subTest(pag_db=pag_db):
                self.assertTrue(str(pag_db) in existing_dbs)

        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(pag_dbs) * 2)

        # check volume and slice .conf files for nabu reconstruction are here
        ini_files = []
        for pag_db in pag_dbs:
            # add volume ini file
            ini_files.append(
                os.path.basename(self.scan.path)
                + "pag_db"
                + str(int(pag_db)).zfill(4)
                + ".cfg"
            )
            # add slice ini file
            # note: plane is only specify for slice.
            ini_files.append(
                os.path.basename(self.scan.path)
                + "slice_pag_000001_db"
                + str(int(pag_db)).zfill(4)
                + "_plane_XY"
                + ".cfg"
            )

        for ini_file in ini_files:
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def testSeveralPagsSeveralSlices(self):
        """behavior when several paganin slices are requested with several
        delta / beta values"""
        slices = (1, 4, 12)
        self.config["tomwer_slices"] = str(slices)
        self.config["phase"]["method"] = "Paganin"
        pag_dbs = [200, 300, 600, 700]
        self.config["phase"]["delta_beta"] = str(pag_dbs)
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            len(slices) * len(pag_dbs) + len(pag_dbs),
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(slices) * len(pag_dbs) + len(pag_dbs))
        # check volume and slice .cfg files for nabu reconstruction are here
        ini_files = []
        for slice_index in slices:
            for pag_db in pag_dbs:
                # add volume ini file
                ini_files.append(
                    os.path.basename(self.scan.path)
                    + "pag_db"
                    + str(pag_db).zfill(4)
                    + ".cfg"
                )
                # add slice ini file
                ini_files.append(
                    os.path.basename(self.scan.path)
                    + "slice_pag_"
                    + str(slice_index).zfill(6)
                    + "_db"
                    + str(pag_db).zfill(4)
                    + "_plane_XY"
                    + ".cfg"
                )

        for ini_file in ini_files:
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def getNabuIniFiles(self):
        return glob(self._nabu_cfg_folder + os.sep + "*.cfg")


class TestTomwerConfForNabu(unittest.TestCase):
    """
    Test the nabu configuration created from tomwer.
    Tomwer can request several slice or several paganin db values.
    We insure we can retrieve the correct nabu configuration for those

    note: the `_interpret_tomwer_configuration` function create each time a
          'None' slice which is the reconstruction of the entire volume
    """

    def setUp(self) -> None:
        self.scan = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        # initial configuration
        self.conf = {"reconstruction": {}}

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def testSlices(self):
        # check if single value
        for data_slice in ("1", NabuSliceMode.MIDDLE):
            with self.subTest(data_slice=data_slice):
                self.conf["tomwer_slices"] = data_slice

                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                # not 1 because return each time the 'None' slice which is
                # the reconstruction of the entire volume
                self.assertEqual(len(sub_confs), 1 + 1)

        # check single values or list of values
        for data_slice in ("[2, 3, 7]", "(6, 9,56)", "5;6; 9"):
            with self.subTest(slice=data_slice):
                self.conf["tomwer_slices"] = data_slice

                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                self.assertEqual(len(sub_confs), 3 + 1)
        # check from:to:step
        self.conf["tomwer_slices"] = "2:10:2"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        self.assertEqual(len(sub_confs), 5 + 1)

    def testPaganinDB(self):
        """Test that tomwer generate one configuration per paganin db"""
        self.conf["phase"] = {}
        # check if single value
        self.conf["phase"]["delta_beta"] = "100"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        # not 1 because return each time the 'None' slice which is
        # the reconstruction of the entire volume
        self.assertEqual(len(sub_confs), 1)

        # check single values or list of values
        for pag_db in ("[100, 200, ]", "(225, 230)", "5;6;"):
            with self.subTest(pag_db=pag_db):
                self.conf["phase"]["delta_beta"] = pag_db
                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                self.assertEqual(len(sub_confs), 2)
        # check from:to:step
        self.conf["phase"]["delta_beta"] = "200:1000:100"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        self.assertEqual(len(sub_confs), 9)


class TestNabuAndAxis(unittest.TestCase):
    """Test that nabu process and axis process are compatible"""

    def setUp(self) -> None:
        self.dataset = "scan_3_"
        self._root_dir = tempfile.mkdtemp()
        data_test_dir = TomwerCIDatasets.get_dataset("edf_datasets/scan_3_")
        scan_folder = os.path.join(self._root_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = EDFTomoScan(scan_folder)
        [os.remove(cfg_file) for cfg_file in glob(self.scan.path + os.sep + "*.cfg")]

        # initial configuration
        self.conf = get_nabu_default_config(output_dir=self.scan.path)

        self.nabu_process = nabu.NabuSlicesTask(
            inputs={
                "data": self.scan,
                "nabu_params": self.conf,
                "dry_run": True,
                "serialize_output_data": False,
            },
        )
        self.scan._axis_params = AxisRP()
        self.nabu_cfg_folder = os.path.join(
            self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
        )
        # make sure no .cfg files already exists
        assert len(glob(self.nabu_cfg_folder + os.sep + "*.cfg")) == 0

    def tearDown(self) -> None:
        shutil.rmtree(self._root_dir)

    def test_output_dir(self):
        self.conf["output"]["location"] = "{scan_basename}/output"
        self.nabu_process.set_configuration(self.conf)
        output_dir = self.scan.scan_basename() + os.sep + "output"
        self.assertFalse(os.path.exists(output_dir))
        self.nabu_process.run()
        self.assertTrue(os.path.exists(output_dir))
        nabu_conf_files = glob(
            output_dir + os.sep + nabu_settings.NABU_CFG_FILE_FOLDER + os.sep + "*.cfg"
        )
        self.assertEqual(len(nabu_conf_files), 1)

    def get_saved_rotation_axis(self, nabu_out_file):
        """return the saved value for 'rotation axis' in the nabu
        configuration file"""
        with open(nabu_out_file, "r") as nabu_file:
            lines = nabu_file
            for line in lines:
                if line.startswith("rotation_axis_position"):
                    return line.replace("\n", "").replace(" ", "").split("=")[-1]
            return None

    def testNoCOR(self):
        """test if axis is executed and no COR are set, nabu won't mind"""
        self.scan._axis_params.set_relative_value(None)
        self.nabu_process.run()

        nabu_conf_files = glob(self.nabu_cfg_folder + os.sep + "*.cfg")
        self.assertEqual(len(nabu_conf_files), 1)
        nabu_rotation_axis_value = self.get_saved_rotation_axis(nabu_conf_files[0])
        self.assertEqual(nabu_rotation_axis_value, "")

    def testCOR(self):
        """test that if cor is computed, cor will be registered for nabu"""
        self.scan._axis_params.set_relative_value(2.3)
        self.nabu_process.run()

        nabu_conf_files = glob(self.nabu_cfg_folder + os.sep + "*.cfg")
        self.assertEqual(len(nabu_conf_files), 1)
        nabu_rotation_axis_value = self.get_saved_rotation_axis(nabu_conf_files[0])
        self.assertEqual(
            nabu_rotation_axis_value,
            str(self.scan._axis_params.value_ref_tomwer + self.scan.dim_1 / 2.0),
        )


def get_nabu_default_config(output_dir):
    """:return a default configuration for nabu"""
    return {
        "phase": {
            "padding_type": "edge",
            "method": "",
            "delta_beta": "100.0",
            "margin": 50,
            "unsharp_sigma": 0,
            "unsharp_coeff": 0,
        },
        "preproc": {
            "ccd_filter_threshold": 0.04,
            "take_logarithm": True,
            "log_min_clip": 1e-06,
            "ccd_filter_enabled": 0,
            "log_max_clip": 10.0,
            "flatfield": 1,
        },
        "dataset": {
            "job_name": "hair_A1_50nm_tomo3_1_",
            "location": "/tmp_14_days/payno/hair_A1_50nm_tomo3_1_",
            "flat_file_prefix": "refHST",
            "projections_subsampling": 1,
            "binning": 1,
            "dark_file_prefix": "dark.edf",
            "binning_z": 1,
        },
        "reconstruction": {
            "positivity_constraint": 1,
            "start_z": 0,
            "end_y": -1,
            "optim_algorithm": "chambolle-pock",
            "enable_halftomo": 0,
            "end_z": -1,
            "iterations": 200,
            "start_x": 0,
            "fbp_filter_type": "ramlak",
            "angle_offset": 0.0,
            "end_x": -1,
            "method": "FBP",
            "angles_file": "",
            "rotation_axis_position": "",
            "start_y": 0,
            "preconditioning_filter": 1,
            "padding_type": "zeros",
            "weight_tv": 0.01,
            "translation_movements_file": "",
            "axis_correction_file": "",
        },
        "output": {"file_format": "hdf5", "location": output_dir},
    }



import logging
import os
import pytest
import numpy
from time import sleep

from silx.gui import qt
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from tomoscan.esrf.volume.tiffvolume import TIFFVolume, has_tifffile

from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.visualization.volumeviewer import VolumeViewer

logging.disable(logging.INFO)


@pytest.mark.skipif(not has_tifffile, reason="tifffile not available")
def test_volume_viewer(
    qtapp,  # noqa F811
    tmp_path,
):
    """
    test the volume viewer setting a scan having an HDF5 volume linked to it
    """

    widget = VolumeViewer(parent=None)
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)

    tmp_dir = tmp_path / "test_volume_viewer"
    tmp_dir.mkdir()

    # step 1 - test setting a scan containing a HDF5Volume
    scan = MockNXtomo(
        scan_path=os.path.join(str(tmp_dir), "myscan"),
        n_proj=20,
        n_ini_proj=20,
        dim=10,
    ).scan
    volume = HDF5Volume(
        file_path=os.path.join(scan.path, "volume.hdf5"),
        data_path="entry",
        data=numpy.random.random(60 * 10 * 10).reshape(60, 10, 10),
    )
    volume.save()

    scan.set_latest_vol_reconstructions(
        [
            volume,
        ]
    )

    widget.setScan(scan)
    assert widget._centralWidget.data() is not None
    widget.clear()
    assert widget._centralWidget.data() is None

    # step 2: test setting a a tiff volume dirrectly
    volume = TIFFVolume(
        folder=os.path.join(tmp_dir, "my_tiff_vol"),
        data=numpy.random.random(60 * 100 * 100).reshape(60, 100, 100),
    )
    volume.save()

    # 2.1 test with the data being in cache
    widget.setVolume(volume=volume)
    assert widget._centralWidget.data() is not None
    widget.clear()
    assert widget._centralWidget.data() is None

    # 2.2 test with the data not being in cache anymore
    volume.clear_cache()

    widget.setVolume(volume=volume)
    # wait_for_processing_finished
    while qt.QApplication.instance().hasPendingEvents():
        qt.QApplication.instance().processEvents()
    sleep(1.2)  # wait for the thread to be processed
    while qt.QApplication.instance().hasPendingEvents():
        qt.QApplication.instance().processEvents()
    # end waiting
    assert widget._centralWidget.data() is not None
